<?php

namespace Drupal\shell;

use Drupal\Component\Utility\Unicode;
use Drupal\Core\Link;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Manage execution of shell commands.
 */
class ShellExec {

  use StringTranslationTrait;

  /**
   * The current request.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $currentRequest;

  /**
   * The current working directory.
   *
   * @var string
   */
  protected $currentDirectory;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * The app root.
   *
   * @var string
   */
  protected $appRoot;

  /**
   * Constructs a ShellExec object.
   *
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The current request.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   The CSRF token manager.
   * @param string $app_root
   *   The app root.
   */
  public function __construct(RequestStack $request_stack, AccountProxyInterface $current_user, $app_root) {
    $this->currentRequest = $request_stack->getCurrentRequest();
    $this->currentUser = $current_user;
    $this->appRoot = $app_root;
  }

  /**
   * Since there are some commands the shell cannot handle correctly,
   * we will capture them and send the user special output.  This function
   * does this.
   *
   * @param String $cmd
   * @return String
   */
  protected function handleSpecialCommand($cmd) {
    // This function will display special information for certain
    // commands which normally cannot be handled through Shell.
    $rtn = "";

    // Take out excess spaces...
    $cmd = str_replace("  ", " ", trim($cmd));
    $cmd = str_replace("  ", " ", $cmd);

    $temp = explode(" ", $cmd);

    $the_rest = "";

    for ($t = 0; $t < count($temp); $t++) {
      if ($t == 0) continue; // skip the first one.
      $the_rest .= $temp[$t] . " ";
    }
    $the_rest = trim($the_rest);

    switch (Unicode::strtolower($temp[0])) {
      case "welcome":
        $rtn = $this->t("<div class='user-command'>Welcome to Shell.<br /></div>");
        $rtn .= $this->t('<strong>System:</strong> @phpuname<br />', ['@phpuname' => php_uname()]);
        $rtn .= $this->t('<strong>PHP:</strong> @phpversion <strong>Web server:</strong> @webserver<br /><br />', ['@phpversion' => phpversion(), '@webserver' => $this->currentRequest->server->get('SERVER_SOFTWARE')]);
        $rtn .= $this->t("Some commands are interpreted by this emulated shell differently than you might expect. To see a list of commands (and for other general help) type <b><u>shelp</u></b>.<br /><br />");
        break;

      case "shelp":
        $rtn .= "=============================<br />
                  Shell Help<br />
                 =============================<br />
                 <b>man <i>command</i></b>  ... " . t("Will provide a link to off-site man page.") . "<br />
                 <b>edit, vi, vim, emacs</b> ... " . t("Will provide a link to a file editor.") . "<br />
                 <br />
                 " . t("Special information:") . "<br />
                 " . t("Be aware that Shell is just an emulator.  All commands are actually
                 being sent through PHP to the server.  As such, interactivity is
                 very limited.  This is why you cannot use vi or emacs.  Any other
                 commands or programs which require interactivity also will not work.
                 You are also limited to what commands your server will allow
                 the web user to run.");
        break;

      case "man":
        $rtn .= "<div class='shell-not-supported'>" . t("This shell emulator cannot correctly view man pages.") . "</div>";
        $rtn .= t("Please click the following link to load your intended man page in a new window:") . "<br />
                  <a href='http://www.geona.net/search?s=manpages.info&q=$the_rest' target='_blank'>
                    http://www.geona.net/search?s=manpages.info&q=$the_rest</a>";
        break;

      case "edit":  // Keep going...
      case "vi":   // keep going...
      case "vim":  // keep going...
      case "emacs":  // Now, let's perform an action...
      $rtn .= "<div class='shell-not-supported'>" . t("This shell emulator cannot correctly edit files, so you must use
                  a special file editor.") . "</div>";
      $rtn .= t("Please click the following link to load the file editor in a new window:") . "<br />";
      $rtn .= Link::createFromRoute($this->t("Click to edit @the_rest", ["@the_rest" => $the_rest]), "shell.file.edit", [], ["query" => ["file" => $the_rest,  "cwd" => $this->getCurrentDirectory()], "attributes" => ["target" => "blank"]])->toString();
      break;

      case "less": // keep going.  Less and more will be treated the same.
      case "more":
        $rtn .= "<div class='shell-not-supported'>" . t("This shell emulator cannot provide the interactivity of less/more, so you must use
                  a special file viewer.") . "</div>";
        $rtn .= t("Please click the following link to load the file viewer in a new window:") . "<br />";
        $rtn .= Link::createFromRoute($this->t("Click to view @the_rest", ["@the_rest" => $the_rest]), "shell.file.view", [], ["query" => ["file" => $the_rest,  "cwd" => $this->getCurrentDirectory()], "attributes" => ["target" => "blank"]])->toString();

        break;
    }

    return $rtn;
  }

  /**
   * Actually handles the command sent by the user by passing it to the server
   * and capturing the output.  This code is drawn partially from bzrundi,
   * and the phpterm project found here: http://phpterm.sourceforge.net/
   *
   * @param String $cmd
   * @return String
   */
  public function handleCommand($cmd) {
    $rtn = "";

    // Make sure current directory is set.
    $this->getCurrentDirectory();

    $cmd = trim($cmd);
    // Since we cannot handle certain commands, we should make sure
    // the user isn't trying to use those commands here.
    /////////  Special commands (which we can't handle) ///////////
    $special_cmd = array("welcome", "shelp", "man", "edit", "vi", "vim", "emacs", "less", "more");
    $temp = explode(" ", Unicode::strtolower($cmd));
    if (in_array($temp[0], $special_cmd)) {
      $rtn = $this->handleSpecialCommand($cmd);
    }
    else {
      // Process the command normally
      $aliases = array('la' => 'ls -la', 'll' => 'ls -lvhF', 'dir' => 'ls');

      if (preg_match('/^[[:blank:]]*cd[[:blank:]]*$/', @$cmd)) {
        // A 'cd' command with no parameters, set Shell current directory to
        // Drupal app root.
        $this->setCurrentDirectory($this->appRoot);
      }
      elseif (preg_match('/^[[:blank:]]*cd[[:blank:]]+([^;]+)$/', @$cmd, $regs)) {
        // The current command is 'cd', which we have to handle as an internal shell command.
        // absolute/relative path ?"
        ($regs[1][0] == '/') ? $new_dir = $regs[1] : $new_dir = $this->getCurrentDirectory() . '/' . $regs[1];

        // cosmetics
        while (strpos($new_dir, '/./') !== FALSE) {
          $new_dir = str_replace('/./', '/', $new_dir);
        }
        while (strpos($new_dir, '//') !== FALSE) {
          $new_dir = str_replace('//', '/', $new_dir);
        }
        while (preg_match('|/\.\.(?!\.)|', $new_dir)) {
          $new_dir = preg_replace('|/?[^/]+/\.\.(?!\.)|', '', $new_dir);
        }

        if (is_dir($new_dir)) {
          $this->setCurrentDirectory($new_dir);
        }
        else {
          $rtn .= "Could not change to: $new_dir\n";
        }
      }
      else {
        // The command is not a 'cd' command, so we execute it after
        // changing the directory and save the output.
        $cwd = getcwd();
        chdir($this->getCurrentDirectory());

        // Alias expansion.
        $length = strcspn(@$cmd, " \t");
        $token = Unicode::substr(@$cmd, 0, $length);
        if (isset($aliases[$token])) {
          $cmd = $aliases[$token] . Unicode::substr($cmd, $length);
        }

        $p = proc_open(@$cmd, array(1 => array('pipe', 'w'), 2 => array('pipe', 'w')), $io);

        // Read output sent to stdout.
        while (!feof($io[1])) {
          $rtn .= htmlspecialchars(fgets($io[1]), ENT_COMPAT, 'UTF-8');
        }
        // Read output sent to stderr.
        while (!feof($io[2])) {
          $rtn .= htmlspecialchars(fgets($io[2]), ENT_COMPAT, 'UTF-8');
        }

        fclose($io[1]);
        fclose($io[2]);
        proc_close($p);

        // Restore the current working directory.
        chdir($cwd);
      }
      $rtn  = "<pre class='shell'>$rtn</pre>";
    }

    $rtn .= "<div class='shell-directory-listing'>" . t("@un%@cwd&gt;", array("@un" => $this->currentUser->getAccountName(), "@cwd" => $this->getCurrentDirectory())) ."</div>";
    return $rtn;
  }

  /**
   * Like it says on the tin, this will perform an "ls" command on the
   * supplied directory, so we can get a list of files.  This is
   * being used by the handle_tab_press function.
   *
   * @param String $dir
   * @return array
   */
  public function getLsOfDirectory($dir) {
    $rtn = array();
    $str = "";
    $p = proc_open(@"ls $dir",
    array(1 => array('pipe', 'w'),
    2 => array('pipe', 'w')), $io);

    // Read output sent to stdout.
    while (!feof($io[1])) {
      $str .= htmlspecialchars(fgets($io[1]), ENT_COMPAT, 'UTF-8');
    }
    // Read output sent to stderr.
    while (!feof($io[2])) {
      $str .= htmlspecialchars(fgets($io[2]), ENT_COMPAT, 'UTF-8');
    }

    fclose($io[1]);
    fclose($io[2]);
    proc_close($p);

    $temp = explode("\n", $str);
    foreach ($temp as $t) {
      $t = trim($t);
      if ($t != "") {
        // If this file is a directory, add a / to the end.
        $is_directory = (is_dir("$dir/$t"))? "/" : "";

        $rtn[] = $t . $is_directory;
      }
    }

    return $rtn;
  }

  /**
   * Handy function from php.net which will give you the file permissions
   * for a particular file.  It will look like it does in linux.
   * Ex: -rw-r--r--
   *
   * @param String $file
   * @return String
   */
  public function getFilePermissions($file) {
    $perms = fileperms($file);

    if (($perms & 0xC000) == 0xC000) {
      // Socket
      $info = 's';
    }
    elseif (($perms & 0xA000) == 0xA000) {
      // Symbolic Link
      $info = 'l';
    }
    elseif (($perms & 0x8000) == 0x8000) {
      // Regular
      $info = '-';
    }
    elseif (($perms & 0x6000) == 0x6000) {
      // Block special
      $info = 'b';
    }
    elseif (($perms & 0x4000) == 0x4000) {
      // Directory
      $info = 'd';
    }
    elseif (($perms & 0x2000) == 0x2000) {
      // Character special
      $info = 'c';
    }
    elseif (($perms & 0x1000) == 0x1000) {
      // FIFO pipe
      $info = 'p';
    }
    else {
      // Unknown
      $info = 'u';
    }

    // Owner
    $info .= (($perms & 0x0100) ? 'r' : '-');
    $info .= (($perms & 0x0080) ? 'w' : '-');
    $info .= (($perms & 0x0040) ?
    (($perms & 0x0800) ? 's' : 'x' ) :
    (($perms & 0x0800) ? 'S' : '-'));

    // Group
    $info .= (($perms & 0x0020) ? 'r' : '-');
    $info .= (($perms & 0x0010) ? 'w' : '-');
    $info .= (($perms & 0x0008) ?
    (($perms & 0x0400) ? 's' : 'x' ) :
    (($perms & 0x0400) ? 'S' : '-'));

    // World
    $info .= (($perms & 0x0004) ? 'r' : '-');
    $info .= (($perms & 0x0002) ? 'w' : '-');
    $info .= (($perms & 0x0001) ?
    (($perms & 0x0200) ? 't' : 'x' ) :
    (($perms & 0x0200) ? 'T' : '-'));

    return $info;
  }

  /**
   * Get the current directory Shell is positioned at.
   *
   * On first call, it is initialized to Drupal root directory.
   *
   * @return string
   *   The current directory Shell is positioned at.
   */
  public function getCurrentDirectory() {
    if ($this->currentDirectory === NULL) {
      $this->setCurrentDirectory($this->appRoot);
    }
    return $this->currentDirectory;
  }

  /**
   * Set the current directory for Shell.
   *
   * @param string $path
   *   A local path.
   *
   * @return $this
   */
  public function setCurrentDirectory($path) {
    $this->currentDirectory = $path;
    return $this;
  }
}
